from __future__ import annotations
from requests import Session
from collections import UserList
from html.parser import unescape
from .utils import kodik_decoder
import re

# aniboom regular expressions (work after unescape response html page)
RE_ANIBOOM = re.compile(r'"hls":"{\\"src\\":\\"(.*\.m3u8)\\"')

# kodik/anivod regular expressions
RE_KODIK_URL = re.compile(r"https://\w+\.\w{2,6}/seria/\d+/\w+/\d{3,4}p")
RE_KODIK_URL_DATA = re.compile(r'iframe.src = "//(.*?)"')
RE_KODIK_VIDEO_TYPE = re.compile(r"go/(\w+)/\d+")
RE_KODIK_VIDEO_ID = re.compile(r"go/\w+/(\d+)")
RE_KODIK_VIDEO_HASH = re.compile(r"go/\w+/\d+/(.*?)/\d+p\?")


class ListObj(UserList):
    """Modified list object"""

    def print_enumerate(self, *args):
        """print elements with getattr names arg. Default invoke __str__ method"""
        if len(self) > 0:
            for i, obj in enumerate(self, 1):
                if args:
                    print(f"[{i}]", *(getattr(obj, arg) for arg in args))
                else:
                    print(f"[{i}]", obj)
        else:
            print("Results not founded!")


class BaseObj:
    """base object for responses"""
    REGEX: dict  # {"attr_name": re.compile(<regular expression>)}

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, str):
                v = int(v) if v.isdigit() else unescape(str(v))
            setattr(self, k, v)

    @classmethod
    def parse(cls, html: str) -> ListObj:
        """class object factory"""
        l_objects = ListObj()
        # generate dict like {attr_name: list(values)}
        results = {k: re.findall(v, html) for k, v in cls.REGEX.items()}
        for values in zip(*results.values()):
            attrs = zip(results.keys(), values)
            # generate objects like {attr_name: attr_value}
            l_objects.append(cls(**dict(attrs)))
        return l_objects


class BaseAnime:
    BASE_URL = "your_base_source_link"
    # mobile user-agent can sometimes gives a chance to bypass the anime title ban
    USER_AGENT = {
        "user-agent":
            "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.114 Mobile Safari/537.36",
            "x-requested-with": "XMLHttpRequest"}
    _instance = None

    def __new__(cls, *args, **kwargs):
        # singleton for correct store session
        if not cls._instance:
            cls._instance = super(BaseAnime, cls).__new__(
                cls, *args, **kwargs)
        return cls._instance

    def __init__(self, session: Session = None):
        if session:
            self.session = session
            self.session.headers.update({"x-requested-with": "XMLHttpRequest"})
        else:
            self.session = Session()
            self.session.headers.update(self.USER_AGENT)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()
        return

    def request(self, method, url, **kwargs):
        return self.session.request(method, url, **kwargs)

    def request_get(self, url, **kwargs):
        return self.request("GET", url, **kwargs)

    def search(self, q: str):
        raise NotImplementedError

    def ongoing(self, *args, **kwargs):
        raise NotImplementedError

    def episodes(self, *args, **kwargs):
        raise NotImplementedError

    def players(self, *args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def _get_kodik_payload(resp: str, referer: str) -> tuple[dict, str]:
        # prepare values for next POST request
        url_data, = re.findall(RE_KODIK_URL_DATA, resp)
        type_, = re.findall(RE_KODIK_VIDEO_TYPE, url_data)
        id_, = re.findall(RE_KODIK_VIDEO_ID, url_data)
        hash_, = re.findall(RE_KODIK_VIDEO_HASH, url_data)
        data = {value.split("=")[0]: value.split("=")[1] for value in url_data.split("?", 1)[1].split("&")}
        data.update({"type": type_, "hash": hash_, "id": id_, "info": {}, "bad_user": True,
                     "ref": referer.rstrip("/")})
        return data, url_data

    def get_kodik_url(self, player_url: str, quality: int = 720, *, referer: str = "") -> str:
        """Get hls url in kodik/anivod player"""
        quality_available = (720, 480, 360, 240, 144)
        if quality not in quality_available:
            quality = 720
        quality = str(quality)

        resp = self.request_get(player_url, headers=self.USER_AGENT.copy().update({"referer": referer}))

        data, url_data = self._get_kodik_payload(resp.text, referer)
        # kodik server regular expr detection
        if not RE_KODIK_URL.match(player_url):
            raise TypeError(
                f"Unknown player balancer. get_kodik_url method support kodik and anivod players\nvideo url: {player_url}")

        url_, = RE_KODIK_URL.findall(player_url)
        url = f"https://{url_}/gvi"
        resp = self.request("POST", url, data=data,
                            headers=self.USER_AGENT.copy().update({"referer": f"https://{url_data}",
                                                                   "orgign": url.replace("/gvi", ""),
                                                                   "accept":
                                                                       "application/json, text/javascript, */*; q=0.01"
                                                                   })).json()["links"]
        video_url = resp["480"][0]["src"]
        video_url = kodik_decoder(video_url).replace("480.mp4", f"{quality}.mp4")
        return video_url

    def get_aniboom_url(self, player_url: str) -> str:
        """get aniboom video"""
        # fix 28 11 2021 request
        r = self.request_get(
            player_url,
            headers={
                "referer": "https://animego.org/",
                "user-agent":
                    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.114 "
                    "Safari/537.36"})

        r = unescape(r.text)
        return re.findall(RE_ANIBOOM, r)[0].replace("\\", "")

    def get_video(self, player_url: str, quality: int = 720, *, referer: str = ""):
        """Return direct video url

        :param Player player: player object
        :return: direct video url
        """
        if "sibnet" in player_url:
            return player_url
        elif "aniboom" in player_url:
            url = self.get_aniboom_url(player_url)
            return url
        elif RE_KODIK_URL.match(player_url):
            url = self.get_kodik_url(player_url, quality, referer=referer)
            return url
        else:
            # catch any players for add in script
            print("Warning!", player_url, "is not supported!")



from __future__ import annotations
from html import unescape
from typing import Union
from anicli_ru.base import ListObj, BaseObj, BaseAnime
import re


class AnimeResult(BaseObj):
    """
    url str: anime url

    title str: title name
    """
    REGEX = {"url": re.compile(r'<a href="(https://animego\.org/anime/.*)" title=".*?">'),
             "title": re.compile('<a href="https://animego\.org/anime/.*" title="(.*?)">')}
    url: str
    title: str

    @property
    def id(self) -> str:
        return self.url.split("-")[-1]

    def __str__(self):
        return f"{self.title}"

    def episodes(self):
        with Anime() as a:
            return a.episodes(result=self)


class Ongoing(BaseObj):
    """
    title: str title name

    num: str episode number

    dub: str dubbing name

    url: str url
    """
    REGEX = {
        "raw_url": re.compile(r'onclick="location\.href=\'(.*?)\'"'),
        "title": re.compile(r'600">(.*?)</span></span></div><div class="ml-3 text-right">'),
        "num":
            re.compile(r'<div class="font-weight-600 text-truncate">(\d+) серия</div><div class="text-gray-dark-6">'),
        "dub": re.compile(r'<div class="text-gray-dark-6">(\(.*?\))</div>'),
        "thumbnail": re.compile(r'"background-image: url\((.*?)\);"></div>')
    }

    title: str
    num: str
    dub: str
    raw_url: str

    thumbnail: str

    @classmethod
    def parse(cls, html: str) -> ListObj:
        ongoings = ListObj()
        # generate dict like {attr_name: list(values)}
        results = {k: re.findall(v, html) for k, v in cls.REGEX.items()}

        for values in zip(*results.values()):
            attrs = zip(results.keys(), values)
            # generate objects like {attr_name: attr_value}
            ongoings.append(cls(**dict((k, v) for k, v in attrs)))

        # shitty sort duplicates (by title and episode num) algorithm
        # but ongoings list contains less than 100 elements guaranty
        sorted_ongoings = ListObj()
        for ongoing in ongoings:
            if ongoing in sorted_ongoings:
                for sorted_ong in sorted_ongoings:
                    if ongoing == sorted_ong:
                        sorted_ong += ongoing
                        break
            else:
                sorted_ongoings.append(ongoing)

        sorted_ongoings.sort(key=lambda k: k.title)
        return sorted_ongoings

    @property
    def url(self):
        return "https://animego.org" + self.raw_url

    @property
    def id(self) -> str:
        return self.url.split("-")[-1]

    def episodes(self):
        with Anime() as a:
            return a.episodes(result=self)

    def __eq__(self, other):
        """return True if title name and episode num equals"""
        return self.num == other.num and self.title == other.title and other.dub not in self.dub

    def __iadd__(self, other):
        """Add dub name in ongoing object with += operator"""
        self.dub += f" {other.dub}"
        return self

    def __add__(self, other):
        """Add dub name in ongoing object with + operator"""
        self.dub += f" {other.dub}"
        return self

    def __str__(self):
        return f"{self.title} {self.num} {self.dub}"


class Episode(BaseObj):
    """
    num: int episode number

    name: str episode name

    id: int episode video id
    """
    REGEX = {"num": re.compile(r'data-episode="(\d+)"'),
             "id": re.compile(r'data-id="(\d+)"'),
             "name": re.compile(r'data-episode-title="(.*)"'),
             }
    num: int
    name: str
    id: int

    def __str__(self):
        return f"{self.name}"

    def player(self):
        with Anime() as a:
            return a.players(self)


class Player(BaseObj):
    """
    dub_id int: dubbing ing

    dub_name str: dubbing name

    player str: video player url
    """
    REGEX = {
        "dub_name":
            re.compile(
                r'data-dubbing="(\d+)"><span class="video-player-toggle-item-name text-underline-hover">\s+(.*)'),
        "player":
            re.compile(
                r'data-player="(.*?)"\s+data-provider="\d+"\s+data-provide-dubbing="(\d+)"'),
        "dub_id": re.compile(r"")
    }
    dub_name: str
    _player: str
    dub_id: int

    @classmethod
    def parse(cls, html: str) -> ListObj:
        l_objects = ListObj()
        # generate dict like {attr_name: list(values)}
        dub_names = re.findall(cls.REGEX["dub_name"], html)  # dub_id, dub_name
        players = re.findall(cls.REGEX["player"], html)  # player_url, dub_id
        for player, dub_id_1 in players:
            p = Player()
            for dub_id_2, dub_name in dub_names:
                # removed check for catching unsupported players
                if dub_id_1 == dub_id_2:
                    p._player, p.dub_name, p.dub_id = player, dub_name, dub_id_1
                    l_objects.append(p)
        return l_objects

    @property
    def url(self) -> str:
        return self._player_prettify(self._player)

    @staticmethod
    def _player_prettify(player: str):
        return "https:" + unescape(player)

    def get_video(self, quality: int = 720, referer: str = "https://animego.org/"):
        with Anime() as a:
            return a.get_video(self.url, quality, referer=referer)

    def __str__(self):
        u = self._player.replace("//", "").split(".", 1)[0]
        return f"{self.dub_name} ({u})"


class Anime(BaseAnime):
    """Anime class parser

    :example:

    ```python
    a = Anime()
    results = a.search("school")
    results.print_enumerate()
    episodes = results[0].episodes()
    ```
    """
    BASE_URL = "https://animego.org"

    # mobile user-agent can sometimes gives a chance to bypass the anime title ban
    USER_AGENT = {"user-agent":
                      "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, "
                      "like Gecko) Chrome/94.0.4606.114 Mobile Safari/537.36",
                  "x-requested-with": "XMLHttpRequest"}

    def search(self, q: str) -> ListObj[AnimeResult]:
        """Get search results

        :param str q: search query
        :return: anime results list
        :rtype: ListObj
        """
        resp = self.request_get(self.BASE_URL + "/search/anime", params={"q": q}).text
        return ListObj(AnimeResult.parse(resp))

    def ongoing(self) -> ListObj[Ongoing]:
        """Get ongoings

        :return: ongoings results list
        :rtype: ListObj
        """
        resp = self.request_get(self.BASE_URL).text
        return Ongoing.parse(resp)

    def episodes(self, result: Union[AnimeResult, Ongoing]) -> ListObj[Episode]:
        """Get available episodes

        :param result: Ongoing or AnimeSearch object
        :return: list available episodes
        :rtype: ListObj
        """
        resp = self.request_get(self.BASE_URL + f"/anime/{result.id}/player?_allow=true").json()["content"]
        return Episode.parse(resp)

    def players(self, episode: Episode) -> ListObj[Player]:
        """Return video players urls

        :param Episode episode: Episode object
        :return: list available players
        :rtype: ListObj[Player]
        """
        resp = self.request_get(self.BASE_URL + "/anime/series", params={"dubbing": 2, "provider": 24,
                                                                         "episode": episode.num,
                                                                         "id": episode.id}).json()["content"]
        players = Player.parse(resp)
        return players



from setuptools import setup
from anicli_ru import __version__

with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='anicli-ru',
    version=__version__,
    packages=['anicli_ru'],
    url='https://github.com/vypivshiy/ani-cli-ru',
    license='GPL-3',
    author='georgiy',
    author_email='',
    python_requires='>=3.6',
    description='anime-ru grabber video api',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['requests'],
)



from __future__ import annotations
from typing import Union, List
from anicli_ru.base import BaseAnime, ListObj, BaseObj
from html.parser import unescape
import re


class AnimeResult(BaseObj):
    """
    url str: anime url

    title str: title name
    """
    REGEX = {"url": re.compile(r'<a class="short-poster img-box" href="(.*?\.html)" data-title=".*?: .*?"'),
             "title": re.compile(r'<a class="short-poster img-box" href=".*?\.html" data-title=".*?: (.*?)"')}
    url: str
    title: str

    def __str__(self):
        return f"{self.title}"

    def episodes(self):
        with Anime() as a:
            return a.episodes(result=self)


class Ongoing(BaseObj):
    """
    title: str title name

    num: str episode number

    url: str url
    """
    REGEX = {
        "raw_url": re.compile(r'<a class="ksupdate_block_list_link" href="(.*?)">.*?</a>'),
        "title": re.compile(r'<a class="ksupdate_block_list_link" href=".*?">(.*?)</a>'),
        "num":
            re.compile(r'<span class="cell cell-2"><a href=".*?">.*?<br>(\d+) .*? </a></span>'),
    }

    title: str
    num: str
    raw_url: str

    @property
    def url(self):
        return "https://animania.online" + self.raw_url

    def episodes(self):
        with Anime() as a:
            return a.episodes(result=self)

    def __str__(self):
        return f"{self.title} {self.num}"


class Player(BaseObj):
    """
    dub_id int: dubbing ing

    dub_name str: dubbing name

    player str: video player url
    """
    REGEX = {}
    dub_name: str
    _player: str
    dub_id: int
    num: int

    @classmethod
    def parse(cls, html: str) -> ListObj:
        raise NotImplementedError("Get <Player> Object from Episode. Ex: Episode().player()")

    @property
    def url(self) -> str:
        return self._player_prettify(self._player)

    @staticmethod
    def _player_prettify(player: str):
        return "https:" + unescape(player)

    def get_video(self, quality: int = 720, referer: str = "https://animania.online/"):
        with Anime() as a:
            return a.get_video(self.url, quality, referer=referer)

    def __str__(self):
        u = self._player.replace("//", "").split(".", 1)[0]
        return f"{self.num} {self.dub_name} ({u})"


class Episode(BaseObj):
    """
    num: int episode number

    name: str episode name

    id: int episode video id
    """
    REGEX = {
        "video_chunks": re.compile(
            r"""(<li id="season\d+" style="display:none;">(<span onclick="kodikSlider\.player\('.*?', this\);"> .*?</span>){1,})"""),
        "dubs": re.compile(r"""onclick="kodikSlider\.season\('(\d+)', this\)" style="display:none;">(.*?)</span>"""),
        "videos": re.compile(r"""<span onclick="kodikSlider\.player\('(.*?)', this\);">"""),
        "num": re.compile(r'data-episode="(\d+)"'),
        "name": re.compile(r'data-episode-title="(.*)"'),
    }
    dub_id: int
    dub_name: str
    count: int
    videos: List[str]

    def __str__(self):
        return f"{self.dub_name} count: {self.count}"

    def player(self) -> ListObj[Player]:
        _players = ListObj()
        for i, videos in enumerate(self.videos, 1):
            p: Player = Player()
            p.dub_id = self.dub_id
            p.dub_name = self.dub_name
            p._player = videos
            p.num = i
            _players.append(p)
        return _players

    @classmethod
    def parse(cls, html: str) -> ListObj:
        videos_chunks = re.findall(cls.REGEX["video_chunks"], html)
        l_obj = ListObj()
        dubs = re.findall(cls.REGEX["dubs"], html)
        videos = [re.findall(cls.REGEX["videos"], chunk[0]) for chunk in videos_chunks]
        for dub_id, dub_name, count, video in zip([int(n[0]) for n in dubs],  # dub id
                                                  [n[1] for n in dubs],  # dub name
                                                  [len(n) for n in videos],  # count videos
                                                  videos):  # videos urls
            l_obj.append(cls(**{"dub_id": dub_id, "dub_name": dub_name, "count": count, "videos": video}))
        return l_obj

    def __eq__(self, other):
        return self.count == other.count

    def __ge__(self, other):
        return self.count >= other.count

    def __le__(self, other):
        return self.count > other.count


class Anime(BaseAnime):
    BASE_URL = "https://animania.online/index.php"

    def search(self, q: str) -> ListObj[AnimeResult]:
        r = self.request_get(self.BASE_URL, params=dict(do="search", subaction="search", story=q))
        return AnimeResult.parse(r.text)

    def ongoing(self) -> ListObj[Ongoing]:
        r = self.request_get(self.BASE_URL)
        return Ongoing.parse(r.text)

    def episodes(self, result: Union[AnimeResult, Ongoing]) -> ListObj[Episode]:
        r = self.request_get(result.url, headers=self.session.headers.copy().update(
            {"Referer": result.url})
                             )
        return Episode.parse(r.text)

    def players(self):
        # get players from episode object
        raise NotImplementedError


if __name__ == '__main__':
    a = Anime()
    ongs = a.ongoing()
    eps = ongs[0].episodes()
    print()


from base64 import b64decode
from os import system


def kodik_decoder(url_encoded: str) -> str:
    """kodik player video url decoder

    :param str url_encoded: encoded url
    :return: decoded video url"""
    url_encoded = url_encoded[::-1]
    if not url_encoded.endswith("=="):
        url_encoded += "=="
    link = b64decode(url_encoded).decode()
    if not link.startswith("https"):
        link = "https:" + link
    return link


def run_player(url: str, player: str = "mpv", **commands) -> None:
    """

    :param url: hls url
    :param player: local video player. Default mpv
    :param commands: send optional commands.
        key-param="value" convert to: --key-param=value or **{"my-key-param":"c"} = --my-key-param=c
    :return:
    """
    if commands:
        commands = " ".join((f'--{k}="{v}"' for k, v in commands.items()))
        system(f"{player} {url} {commands}")
    else:
        system(f"{player} {url}")


def is_aniboom(url: str):
    """Костыль, который проверяет принадлежность видео к этому балансеру"""
    return "aniboom" in url

